@@ -5,10 +5,9 @@
  Alan Ott
  Signal 11 Software
 
- 8/22/2009
- Linux Version - 6/2/2009
+ libusb/hidapi Team
 
- Copyright 2009, All Rights Reserved.
+ Copyright 2022, All Rights Reserved.
 
  At the discretion of the user of this library,
  this software may be licensed under the terms of the
@@ -62,27 +61,18 @@
 #endif
 
 
-/* USB HID device property names */
-const char *device_string_names[] = {
-	"manufacturer",
-	"product",
-	"serial",
-};
-
-/* Symbolic names for the properties above */
-enum device_string_id {
-	DEVICE_STRING_MANUFACTURER,
-	DEVICE_STRING_PRODUCT,
-	DEVICE_STRING_SERIAL,
-
-	DEVICE_STRING_COUNT,
-};
+// HIDIOCGINPUT is not defined in Linux kernel headers < 5.11.
+// This definition is from hidraw.h in Linux >= 5.11.
+// https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=f43d3870cafa2a0f3854c1819c8385733db8f9ae
+#ifndef HIDIOCGINPUT
+#define HIDIOCGINPUT(len)    _IOC(_IOC_WRITE|_IOC_READ, 'H', 0x0A, len)
+#endif
 
 struct hid_device_ {
 	int device_handle;
 	int blocking;
-	int uses_numbered_reports;
 	wchar_t *last_error_str;
+	struct hid_device_info* device_info;
 };
 
 static struct hid_api_version api_version = {
@@ -91,17 +81,20 @@ static struct hid_api_version api_version = {
 	.patch = HID_API_VERSION_PATCH
 };
 
-/* Global error message that is not specific to a device, e.g. for
-   hid_open(). It is thread-local like errno. */
-__thread wchar_t *last_global_error_str = NULL;
+static wchar_t *last_global_error_str = NULL;
+
 
 static hid_device *new_hid_device(void)
 {
 	hid_device *dev = (hid_device*) calloc(1, sizeof(hid_device));
+	if (dev == NULL) {
+		return NULL;
+	}
+
 	dev->device_handle = -1;
 	dev->blocking = 1;
-	dev->uses_numbered_reports = 0;
 	dev->last_error_str = NULL;
+	dev->device_info = NULL;
 
 	return dev;
 }
@@ -118,6 +111,10 @@ static wchar_t *utf8_to_wchar_t(const char *utf8)
 			return wcsdup(L"");
 		}
 		ret = (wchar_t*) calloc(wlen+1, sizeof(wchar_t));
+		if (ret == NULL) {
+			/* as much as we can do at this point */
+			return NULL;
+		}
 		mbstowcs(ret, utf8, wlen+1);
 		ret[wlen] = 0x0000;
 	}
@@ -126,6 +123,25 @@ static wchar_t *utf8_to_wchar_t(const char *utf8)
 }
 
 
+/* Makes a copy of the given error message (and decoded according to the
+ * currently locale) into the wide string pointer pointed by error_str.
+ * The last stored error string is freed.
+ * Use register_error_str(NULL) to free the error message completely. */
+static void register_error_str(wchar_t **error_str, const char *msg)
+{
+	free(*error_str);
+	*error_str = utf8_to_wchar_t(msg);
+}
+
+/* Semilar to register_error_str, but allows passing a format string with va_list args into this function. */
+static void register_error_str_vformat(wchar_t **error_str, const char *format, va_list args)
+{
+	char msg[256];
+	vsnprintf(msg, sizeof(msg), format, args);
+
+	register_error_str(error_str, msg);
+}
+
 /* Set the last global error to be reported by hid_error(NULL).
  * The given error message will be copied (and decoded according to the
  * currently locale, so do not pass in string constants).
@@ -133,51 +149,35 @@ static wchar_t *utf8_to_wchar_t(const char *utf8)
  * Use register_global_error(NULL) to indicate "no error". */
 static void register_global_error(const char *msg)
 {
-	if (last_global_error_str)
-		free(last_global_error_str);
-
-	last_global_error_str = utf8_to_wchar_t(msg);
+	register_error_str(&last_global_error_str, msg);
 }
 
-/* See register_global_error, but you can pass a format string into this function. */
+/* Similar to register_global_error, but allows passing a format string into this function. */
 static void register_global_error_format(const char *format, ...)
 {
 	va_list args;
 	va_start(args, format);
-
-	char msg[100];
-	vsnprintf(msg, sizeof(msg), format, args);
-
+	register_error_str_vformat(&last_global_error_str, format, args);
 	va_end(args);
-
-	register_global_error(msg);
 }
 
-/* Set the last error for a device to be reported by hid_error(device).
+/* Set the last error for a device to be reported by hid_error(dev).
  * The given error message will be copied (and decoded according to the
  * currently locale, so do not pass in string constants).
- * The last stored global error message is freed.
- * Use register_device_error(device, NULL) to indicate "no error". */
+ * The last stored device error message is freed.
+ * Use register_device_error(dev, NULL) to indicate "no error". */
 static void register_device_error(hid_device *dev, const char *msg)
 {
-	if (dev->last_error_str)
-		free(dev->last_error_str);
-
-	dev->last_error_str = utf8_to_wchar_t(msg);
+	register_error_str(&dev->last_error_str, msg);
 }
 
-/* See register_device_error, but you can pass a format string into this function. */
+/* Similar to register_device_error, but you can pass a format string into this function. */
 static void register_device_error_format(hid_device *dev, const char *format, ...)
 {
 	va_list args;
 	va_start(args, format);
-
-	char msg[100];
-	vsnprintf(msg, sizeof(msg), format, args);
-
+	register_error_str_vformat(&dev->last_error_str, format, args);
 	va_end(args);
-
-	register_device_error(dev, msg);
 }
 
 /* Get an attribute value from a udev_device and return it as a whar_t
@@ -244,34 +244,6 @@ static int get_hid_item_size(__u8 *report_descriptor, unsigned int pos, __u32 si
 	return 0;
 }
 
-/* uses_numbered_reports() returns 1 if report_descriptor describes a device
-   which contains numbered reports. */
-static int uses_numbered_reports(__u8 *report_descriptor, __u32 size) {
-	unsigned int i = 0;
-	int data_len, key_size;
-
-	while (i < size) {
-		int key = report_descriptor[i];
-
-		/* Check for the Report ID key */
-		if (key == 0x85/*Report ID*/) {
-			/* This device has a Report ID, which means it uses
-			   numbered reports. */
-			return 1;
-		}
-
-		/* Determine data_len and key_size */
-		if (!get_hid_item_size(report_descriptor, i, size, &data_len, &key_size))
-			return 0; /* malformed report */
-
-		/* Skip over this key and it's associated data */
-		i += data_len + key_size;
-	}
-
-	/* Didn't find a Report ID key. Device doesn't use numbered reports. */
-	return 0;
-}
-
 /*
  * Get bytes from a HID Report Descriptor.
  * Only call with a num_bytes of 0, 1, 2, or 4.
@@ -361,7 +333,7 @@ static int get_next_hid_usage(__u8 *report_descriptor, __u32 size, unsigned int
 			break;
 		}
 
-		/* Skip over this key and it's associated data */
+		/* Skip over this key and its associated data */
 		*pos += data_len + key_size;
 
 		/* Return usage pair */
@@ -386,7 +358,7 @@ static int get_hid_report_descriptor(const char *rpt_path, struct hidraw_report_
 	int rpt_handle;
 	ssize_t res;
 
-	rpt_handle = open(rpt_path, O_RDONLY);
+	rpt_handle = open(rpt_path, O_RDONLY | O_CLOEXEC);
 	if (rpt_handle < 0) {
 		register_global_error_format("open failed (%s): %s", rpt_path, strerror(errno));
 		return -1;
@@ -409,6 +381,7 @@ static int get_hid_report_descriptor(const char *rpt_path, struct hidraw_report_
 	return (int) res;
 }
 
+/* return size of the descriptor, or -1 on failure */
 static int get_hid_report_descriptor_from_sysfs(const char *sysfs_path, struct hidraw_report_descriptor *rpt_desc)
 {
 	int res = -1;
@@ -423,16 +396,106 @@ static int get_hid_report_descriptor_from_sysfs(const char *sysfs_path, struct h
 	return res;
 }
 
+/* return non-zero if successfully parsed */
+static int parse_hid_vid_pid_from_uevent(const char *uevent, unsigned *bus_type, unsigned short *vendor_id, unsigned short *product_id)
+{
+	char tmp[1024];
+	size_t uevent_len = strlen(uevent);
+	if (uevent_len > sizeof(tmp) - 1)
+		uevent_len = sizeof(tmp) - 1;
+	memcpy(tmp, uevent, uevent_len);
+	tmp[uevent_len] = '\0';
+
+	char *saveptr = NULL;
+	char *line;
+	char *key;
+	char *value;
+
+	line = strtok_r(tmp, "\n", &saveptr);
+	while (line != NULL) {
+		/* line: "KEY=value" */
+		key = line;
+		value = strchr(line, '=');
+		if (!value) {
+			goto next_line;
+		}
+		*value = '\0';
+		value++;
+
+		if (strcmp(key, "HID_ID") == 0) {
+			/**
+			 *        type vendor   product
+			 * HID_ID=0003:000005AC:00008242
+			 **/
+			int ret = sscanf(value, "%x:%hx:%hx", bus_type, vendor_id, product_id);
+			if (ret == 3) {
+				return 1;
+			}
+		}
+
+next_line:
+		line = strtok_r(NULL, "\n", &saveptr);
+	}
+
+	register_global_error("Couldn't find/parse HID_ID");
+	return 0;
+}
+
+/* return non-zero if successfully parsed */
+static int parse_hid_vid_pid_from_uevent_path(const char *uevent_path, unsigned *bus_type, unsigned short *vendor_id, unsigned short *product_id)
+{
+	int handle;
+	ssize_t res;
+
+	handle = open(uevent_path, O_RDONLY | O_CLOEXEC);
+	if (handle < 0) {
+		register_global_error_format("open failed (%s): %s", uevent_path, strerror(errno));
+		return 0;
+	}
+
+	char buf[1024];
+	res = read(handle, buf, sizeof(buf));
+	close(handle);
+
+	if (res < 0) {
+		register_global_error_format("read failed (%s): %s", uevent_path, strerror(errno));
+		return 0;
+	}
+
+	buf[res] = '\0';
+	return parse_hid_vid_pid_from_uevent(buf, bus_type, vendor_id, product_id);
+}
+
+/* return non-zero if successfully read/parsed */
+static int parse_hid_vid_pid_from_sysfs(const char *sysfs_path, unsigned *bus_type, unsigned short *vendor_id, unsigned short *product_id)
+{
+	int res = 0;
+	/* Construct <sysfs_path>/device/uevent */
+	size_t uevent_path_len = strlen(sysfs_path) + 14 + 1;
+	char* uevent_path = (char*) calloc(1, uevent_path_len);
+	snprintf(uevent_path, uevent_path_len, "%s/device/uevent", sysfs_path);
+
+	res = parse_hid_vid_pid_from_uevent_path(uevent_path, bus_type, vendor_id, product_id);
+	free(uevent_path);
+
+	return res;
+}
+
 /*
  * The caller is responsible for free()ing the (newly-allocated) character
  * strings pointed to by serial_number_utf8 and product_name_utf8 after use.
  */
-static int
-parse_uevent_info(const char *uevent, unsigned *bus_type,
+static int parse_uevent_info(const char *uevent, unsigned *bus_type,
 	unsigned short *vendor_id, unsigned short *product_id,
 	char **serial_number_utf8, char **product_name_utf8)
 {
-	char *tmp = strdup(uevent);
+	char tmp[1024];
+	size_t uevent_len = strlen(uevent);
+	if (uevent_len > sizeof(tmp) - 1)
+		uevent_len = sizeof(tmp) - 1;
+	memcpy(tmp, uevent, uevent_len);
+	tmp[uevent_len] = '\0';
+
 	char *saveptr = NULL;
 	char *line;
 	char *key;
@@ -476,109 +539,203 @@ next_line:
 		line = strtok_r(NULL, "\n", &saveptr);
 	}
 
-	free(tmp);
 	return (found_id && found_name && found_serial);
 }
 
 
-static int get_device_string(hid_device *dev, enum device_string_id key, wchar_t *string, size_t maxlen)
+static struct hid_device_info * create_device_info_for_device(struct udev_device *raw_dev)
 {
-	struct udev *udev;
-	struct udev_device *udev_dev, *parent, *hid_dev;
-	struct stat s;
-	int ret = -1;
+	struct hid_device_info *root = NULL;
+	struct hid_device_info *cur_dev = NULL;
+
+	const char *sysfs_path;
+	const char *dev_path;
+	const char *str;
+	struct udev_device *hid_dev; /* The device's HID udev node. */
+	struct udev_device *usb_dev; /* The device's USB udev node. */
+	struct udev_device *intf_dev; /* The device's interface (in the USB sense). */
+	unsigned short dev_vid;
+	unsigned short dev_pid;
 	char *serial_number_utf8 = NULL;
 	char *product_name_utf8 = NULL;
+	unsigned bus_type;
+	int result;
+	struct hidraw_report_descriptor report_desc;
 
-	/* Create the udev object */
-	udev = udev_new();
-	if (!udev) {
-		register_global_error("Couldn't create udev context");
-		return -1;
+	sysfs_path = udev_device_get_syspath(raw_dev);
+	dev_path = udev_device_get_devnode(raw_dev);
+
+	hid_dev = udev_device_get_parent_with_subsystem_devtype(
+		raw_dev,
+		"hid",
+		NULL);
+
+	if (!hid_dev) {
+		/* Unable to find parent hid device. */
+		goto end;
 	}
 
-	/* Get the dev_t (major/minor numbers) from the file handle. */
-	ret = fstat(dev->device_handle, &s);
-	if (-1 == ret)
-		return ret;
-	/* Open a udev device from the dev_t. 'c' means character device. */
-	udev_dev = udev_device_new_from_devnum(udev, 'c', s.st_rdev);
-	if (udev_dev) {
-		hid_dev = udev_device_get_parent_with_subsystem_devtype(
-			udev_dev,
-			"hid",
-			NULL);
-		if (hid_dev) {
-			unsigned short dev_vid;
-			unsigned short dev_pid;
-			unsigned bus_type;
-			size_t retm;
+	result = parse_uevent_info(
+		udev_device_get_sysattr_value(hid_dev, "uevent"),
+		&bus_type,
+		&dev_vid,
+		&dev_pid,
+		&serial_number_utf8,
+		&product_name_utf8);
 
-			ret = parse_uevent_info(
-			           udev_device_get_sysattr_value(hid_dev, "uevent"),
-			           &bus_type,
-			           &dev_vid,
-			           &dev_pid,
-			           &serial_number_utf8,
-			           &product_name_utf8);
+	if (!result) {
+		/* parse_uevent_info() failed for at least one field. */
+		goto end;
+	}
 
-			/* Standard USB device */
-			if (bus_type == BUS_USB) {
-				/* This is a USB device. Find its parent USB Device node. */
-				parent = udev_device_get_parent_with_subsystem_devtype(
-					   udev_dev,
-					   "usb",
-					   "usb_device");
-				if (parent) {
-					const char *str;
-					const char *key_str = NULL;
+	/* Filter out unhandled devices right away */
+	switch (bus_type) {
+		case BUS_BLUETOOTH:
+		case BUS_I2C:
+		case BUS_USB:
+		case BUS_SPI:
+			break;
 
-					if (key >= 0 && key < DEVICE_STRING_COUNT) {
-						key_str = device_string_names[key];
-					} else {
-						ret = -1;
-						goto end;
-					}
+		default:
+			goto end;
+	}
 
-					str = udev_device_get_sysattr_value(parent, key_str);
-					if (str) {
-						/* Convert the string from UTF-8 to wchar_t */
-						retm = mbstowcs(string, str, maxlen);
-						ret = (retm == (size_t)-1)? -1: 0;
-					}
+	/* Create the record. */
+	root = (struct hid_device_info*) calloc(1, sizeof(struct hid_device_info));
+	if (!root)
+		goto end;
 
-					/* USB information parsed */
-					goto end;
-				}
-				else {
-					/* Correctly handled below */
-				}
+	cur_dev = root;
+
+	/* Fill out the record */
+	cur_dev->next = NULL;
+	cur_dev->path = dev_path? strdup(dev_path): NULL;
+
+	/* VID/PID */
+	cur_dev->vendor_id = dev_vid;
+	cur_dev->product_id = dev_pid;
+
+	/* Serial Number */
+	cur_dev->serial_number = utf8_to_wchar_t(serial_number_utf8);
+
+	/* Release Number */
+	cur_dev->release_number = 0x0;
+
+	/* Interface Number */
+	cur_dev->interface_number = -1;
+
+	switch (bus_type) {
+		case BUS_USB:
+			/* The device pointed to by raw_dev contains information about
+				the hidraw device. In order to get information about the
+				USB device, get the parent device with the
+				subsystem/devtype pair of "usb"/"usb_device". This will
+				be several levels up the tree, but the function will find
+				it. */
+			usb_dev = udev_device_get_parent_with_subsystem_devtype(
+					raw_dev,
+					"usb",
+					"usb_device");
+
+			/* uhid USB devices
+			 * Since this is a virtual hid interface, no USB information will
+			 * be available. */
+			if (!usb_dev) {
+				/* Manufacturer and Product strings */
+				cur_dev->manufacturer_string = wcsdup(L"");
+				cur_dev->product_string = utf8_to_wchar_t(product_name_utf8);
+				break;
 			}
 
-			/* USB information not available (uhid) or another type of HID bus */
-			switch (bus_type) {
-				case BUS_BLUETOOTH:
-				case BUS_I2C:
-				case BUS_USB:
-					switch (key) {
-						case DEVICE_STRING_MANUFACTURER:
-							wcsncpy(string, L"", maxlen);
-							ret = 0;
-							break;
-						case DEVICE_STRING_PRODUCT:
-							retm = mbstowcs(string, product_name_utf8, maxlen);
-							ret = (retm == (size_t)-1)? -1: 0;
-							break;
-						case DEVICE_STRING_SERIAL:
-							retm = mbstowcs(string, serial_number_utf8, maxlen);
-							ret = (retm == (size_t)-1)? -1: 0;
-							break;
-						case DEVICE_STRING_COUNT:
-						default:
-							ret = -1;
-							break;
-					}
+			cur_dev->manufacturer_string = copy_udev_string(usb_dev, "manufacturer");
+			cur_dev->product_string = copy_udev_string(usb_dev, "product");
+
+			cur_dev->bus_type = HID_API_BUS_USB;
+
+			str = udev_device_get_sysattr_value(usb_dev, "bcdDevice");
+			cur_dev->release_number = (str)? strtol(str, NULL, 16): 0x0;
+
+			/* Get a handle to the interface's udev node. */
+			intf_dev = udev_device_get_parent_with_subsystem_devtype(
+					raw_dev,
+					"usb",
+					"usb_interface");
+			if (intf_dev) {
+				str = udev_device_get_sysattr_value(intf_dev, "bInterfaceNumber");
+				cur_dev->interface_number = (str)? strtol(str, NULL, 16): -1;
 			}
+
+			break;
+
+		case BUS_BLUETOOTH:
+			cur_dev->manufacturer_string = wcsdup(L"");
+			cur_dev->product_string = utf8_to_wchar_t(product_name_utf8);
+
+			cur_dev->bus_type = HID_API_BUS_BLUETOOTH;
+
+			break;
+		case BUS_I2C:
+			cur_dev->manufacturer_string = wcsdup(L"");
+			cur_dev->product_string = utf8_to_wchar_t(product_name_utf8);
+
+			cur_dev->bus_type = HID_API_BUS_I2C;
+
+			break;
+
+		case BUS_SPI:
+			cur_dev->manufacturer_string = wcsdup(L"");
+			cur_dev->product_string = utf8_to_wchar_t(product_name_utf8);
+
+			cur_dev->bus_type = HID_API_BUS_SPI;
+
+			break;
+
+		default:
+			/* Unknown device type - this should never happen, as we
+			 * check for USB and Bluetooth devices above */
+			break;
+	}
+
+	/* Usage Page and Usage */
+	result = get_hid_report_descriptor_from_sysfs(sysfs_path, &report_desc);
+	if (result >= 0) {
+		unsigned short page = 0, usage = 0;
+		unsigned int pos = 0;
+		/*
+		 * Parse the first usage and usage page
+		 * out of the report descriptor.
+		 */
+		if (!get_next_hid_usage(report_desc.value, report_desc.size, &pos, &page, &usage)) {
+			cur_dev->usage_page = page;
+			cur_dev->usage = usage;
+		}
+
+		/*
+		 * Parse any additional usage and usage pages
+		 * out of the report descriptor.
+		 */
+		while (!get_next_hid_usage(report_desc.value, report_desc.size, &pos, &page, &usage)) {
+			/* Create new record for additional usage pairs */
+			struct hid_device_info *tmp = (struct hid_device_info*) calloc(1, sizeof(struct hid_device_info));
+			struct hid_device_info *prev_dev = cur_dev;
+
+			if (!tmp)
+				continue;
+			cur_dev->next = tmp;
+			cur_dev = tmp;
+
+			/* Update fields */
+			cur_dev->path = dev_path? strdup(dev_path): NULL;
+			cur_dev->vendor_id = dev_vid;
+			cur_dev->product_id = dev_pid;
+			cur_dev->serial_number = prev_dev->serial_number? wcsdup(prev_dev->serial_number): NULL;
+			cur_dev->release_number = prev_dev->release_number;
+			cur_dev->interface_number = prev_dev->interface_number;
+			cur_dev->manufacturer_string = prev_dev->manufacturer_string? wcsdup(prev_dev->manufacturer_string): NULL;
+			cur_dev->product_string = prev_dev->product_string? wcsdup(prev_dev->product_string): NULL;
+			cur_dev->usage_page = page;
+			cur_dev->usage = usage;
+			cur_dev->bus_type = prev_dev->bus_type;
 		}
 	}
 
@@ -586,12 +743,47 @@ end:
 	free(serial_number_utf8);
 	free(product_name_utf8);
 
+	return root;
+}
+
+static struct hid_device_info * create_device_info_for_hid_device(hid_device *dev) {
+	struct udev *udev;
+	struct udev_device *udev_dev;
+	struct stat s;
+	int ret = -1;
+	struct hid_device_info *root = NULL;
+
+	register_device_error(dev, NULL);
+
+	/* Get the dev_t (major/minor numbers) from the file handle. */
+	ret = fstat(dev->device_handle, &s);
+	if (-1 == ret) {
+		register_device_error(dev, "Failed to stat device handle");
+		return NULL;
+	}
+
+	/* Create the udev object */
+	udev = udev_new();
+	if (!udev) {
+		register_device_error(dev, "Couldn't create udev context");
+		return NULL;
+	}
+
+	/* Open a udev device from the dev_t. 'c' means character device. */
+	udev_dev = udev_device_new_from_devnum(udev, 'c', s.st_rdev);
+	if (udev_dev) {
+		root = create_device_info_for_device(udev_dev);
+	}
+
+	if (!root) {
+		/* TODO: have a better error reporting via create_device_info_for_device */
+		register_device_error(dev, "Couldn't create hid_device_info");
+	}
+
 	udev_device_unref(udev_dev);
-	/* parent and hid_dev don't need to be (and can't be) unref'd.
-	   I'm not sure why, but they'll throw double-free() errors. */
 	udev_unref(udev);
 
-	return ret;
+	return root;
 }
 
 HID_API_EXPORT const struct hid_api_version* HID_API_CALL hid_version()
@@ -608,6 +800,9 @@ int HID_API_EXPORT hid_init(void)
 {
 	const char *locale;
 
+	/* indicate no error */
+	register_global_error(NULL);
+
 	/* Set the locale if it's not set. */
 	locale = setlocale(LC_CTYPE, NULL);
 	if (!locale)
@@ -624,7 +819,6 @@ int HID_API_EXPORT hid_exit(void)
 	return 0;
 }
 
-
 struct hid_device_info  HID_API_EXPORT *hid_enumerate(unsigned short vendor_id, unsigned short product_id)
 {
 	struct udev *udev;
@@ -633,9 +827,9 @@ struct hid_device_info  HID_API_EXPORT *hid_enumerate(unsigned short vendor_id,
 
 	struct hid_device_info *root = NULL; /* return object */
 	struct hid_device_info *cur_dev = NULL;
-	struct hid_device_info *prev_dev = NULL; /* previous device */
 
 	hid_init();
+	/* register_global_error: global error is reset by hid_init */
 
 	/* Create the udev object */
 	udev = udev_new();
@@ -653,202 +847,62 @@ struct hid_device_info  HID_API_EXPORT *hid_enumerate(unsigned short vendor_id,
 	   create a udev_device record for it */
 	udev_list_entry_foreach(dev_list_entry, devices) {
 		const char *sysfs_path;
-		const char *dev_path;
-		const char *str;
+		unsigned short dev_vid = 0;
+		unsigned short dev_pid = 0;
+		unsigned bus_type = 0;
 		struct udev_device *raw_dev; /* The device's hidraw udev node. */
-		struct udev_device *hid_dev; /* The device's HID udev node. */
-		struct udev_device *usb_dev; /* The device's USB udev node. */
-		struct udev_device *intf_dev; /* The device's interface (in the USB sense). */
-		unsigned short dev_vid;
-		unsigned short dev_pid;
-		char *serial_number_utf8 = NULL;
-		char *product_name_utf8 = NULL;
-		unsigned bus_type;
-		int result;
-		struct hidraw_report_descriptor report_desc;
+		struct hid_device_info * tmp;
 
 		/* Get the filename of the /sys entry for the device
 		   and create a udev_device object (dev) representing it */
 		sysfs_path = udev_list_entry_get_name(dev_list_entry);
+		if (!sysfs_path)
+			continue;
+
+		if (vendor_id != 0 || product_id != 0) {
+			if (!parse_hid_vid_pid_from_sysfs(sysfs_path, &bus_type, &dev_vid, &dev_pid))
+				continue;
+
+			if (vendor_id != 0 && vendor_id != dev_vid)
+				continue;
+			if (product_id != 0 && product_id != dev_pid)
+				continue;
+		}
+
 		raw_dev = udev_device_new_from_syspath(udev, sysfs_path);
-		dev_path = udev_device_get_devnode(raw_dev);
+		if (!raw_dev)
+			continue;
 
-		hid_dev = udev_device_get_parent_with_subsystem_devtype(
-			raw_dev,
-			"hid",
-			NULL);
-
-		if (!hid_dev) {
-			/* Unable to find parent hid device. */
-			goto next;
-		}
-
-		result = parse_uevent_info(
-			udev_device_get_sysattr_value(hid_dev, "uevent"),
-			&bus_type,
-			&dev_vid,
-			&dev_pid,
-			&serial_number_utf8,
-			&product_name_utf8);
-
-		if (!result) {
-			/* parse_uevent_info() failed for at least one field. */
-			goto next;
-		}
-
-		/* Filter out unhandled devices right away */
-		switch (bus_type) {
-			case BUS_BLUETOOTH:
-			case BUS_I2C:
-			case BUS_USB:
-				break;
-
-			default:
-				goto next;
-		}
-
-		/* Check the VID/PID against the arguments */
-		if ((vendor_id == 0x0 || vendor_id == dev_vid) &&
-		    (product_id == 0x0 || product_id == dev_pid)) {
-			struct hid_device_info *tmp;
-
-			/* VID/PID match. Create the record. */
-			tmp = (struct hid_device_info*) calloc(1, sizeof(struct hid_device_info));
+		tmp = create_device_info_for_device(raw_dev);
+		if (tmp) {
 			if (cur_dev) {
 				cur_dev->next = tmp;
 			}
 			else {
 				root = tmp;
 			}
-			prev_dev = cur_dev;
 			cur_dev = tmp;
 
-			/* Fill out the record */
-			cur_dev->next = NULL;
-			cur_dev->path = dev_path? strdup(dev_path): NULL;
-
-			/* VID/PID */
-			cur_dev->vendor_id = dev_vid;
-			cur_dev->product_id = dev_pid;
-
-			/* Serial Number */
-			cur_dev->serial_number = utf8_to_wchar_t(serial_number_utf8);
-
-			/* Release Number */
-			cur_dev->release_number = 0x0;
-
-			/* Interface Number */
-			cur_dev->interface_number = -1;
-
-			switch (bus_type) {
-				case BUS_USB:
-					/* The device pointed to by raw_dev contains information about
-					   the hidraw device. In order to get information about the
-					   USB device, get the parent device with the
-					   subsystem/devtype pair of "usb"/"usb_device". This will
-					   be several levels up the tree, but the function will find
-					   it. */
-					usb_dev = udev_device_get_parent_with_subsystem_devtype(
-							raw_dev,
-							"usb",
-							"usb_device");
-
-					/* uhid USB devices
-					   Since this is a virtual hid interface, no USB information will
-					   be available. */
-					if (!usb_dev) {
-						/* Manufacturer and Product strings */
-						cur_dev->manufacturer_string = wcsdup(L"");
-						cur_dev->product_string = utf8_to_wchar_t(product_name_utf8);
-						break;
-					}
-
-					/* Manufacturer and Product strings */
-					cur_dev->manufacturer_string = copy_udev_string(usb_dev, device_string_names[DEVICE_STRING_MANUFACTURER]);
-					cur_dev->product_string = copy_udev_string(usb_dev, device_string_names[DEVICE_STRING_PRODUCT]);
-
-					/* Release Number */
-					str = udev_device_get_sysattr_value(usb_dev, "bcdDevice");
-					cur_dev->release_number = (str)? strtol(str, NULL, 16): 0x0;
-
-					/* Get a handle to the interface's udev node. */
-					intf_dev = udev_device_get_parent_with_subsystem_devtype(
-							raw_dev,
-							"usb",
-							"usb_interface");
-					if (intf_dev) {
-						str = udev_device_get_sysattr_value(intf_dev, "bInterfaceNumber");
-						cur_dev->interface_number = (str)? strtol(str, NULL, 16): -1;
-					}
-
-					break;
-
-				case BUS_BLUETOOTH:
-				case BUS_I2C:
-					/* Manufacturer and Product strings */
-					cur_dev->manufacturer_string = wcsdup(L"");
-					cur_dev->product_string = utf8_to_wchar_t(product_name_utf8);
-
-					break;
-
-				default:
-					/* Unknown device type - this should never happen, as we
-					 * check for USB and Bluetooth devices above */
-					break;
-			}
-
-			/* Usage Page and Usage */
-			result = get_hid_report_descriptor_from_sysfs(sysfs_path, &report_desc);
-			if (result >= 0) {
-				unsigned short page = 0, usage = 0;
-				unsigned int pos = 0;
-				/*
-				 * Parse the first usage and usage page
-				 * out of the report descriptor.
-				 */
-				if (!get_next_hid_usage(report_desc.value, report_desc.size, &pos, &page, &usage)) {
-					cur_dev->usage_page = page;
-					cur_dev->usage = usage;
-				}
-
-				/*
-				 * Parse any additional usage and usage pages
-				 * out of the report descriptor.
-				 */
-				while (!get_next_hid_usage(report_desc.value, report_desc.size, &pos, &page, &usage)) {
-					/* Create new record for additional usage pairs */
-					tmp = (struct hid_device_info*) calloc(1, sizeof(struct hid_device_info));
-					cur_dev->next = tmp;
-					prev_dev = cur_dev;
-					cur_dev = tmp;
-
-					/* Update fields */
-					cur_dev->path = strdup(dev_path);
-					cur_dev->vendor_id = dev_vid;
-					cur_dev->product_id = dev_pid;
-					cur_dev->serial_number = prev_dev->serial_number? wcsdup(prev_dev->serial_number): NULL;
-					cur_dev->release_number = prev_dev->release_number;
-					cur_dev->interface_number = prev_dev->interface_number;
-					cur_dev->manufacturer_string = prev_dev->manufacturer_string? wcsdup(prev_dev->manufacturer_string): NULL;
-					cur_dev->product_string = prev_dev->product_string? wcsdup(prev_dev->product_string): NULL;
-					cur_dev->usage_page = page;
-					cur_dev->usage = usage;
-				}
+			/* move the pointer to the tail of returnd list */
+			while (cur_dev->next != NULL) {
+				cur_dev = cur_dev->next;
 			}
 		}
 
-	next:
-		free(serial_number_utf8);
-		free(product_name_utf8);
 		udev_device_unref(raw_dev);
-		/* hid_dev, usb_dev and intf_dev don't need to be (and can't be)
-		   unref()d.  It will cause a double-free() error.  I'm not
-		   sure why.  */
 	}
 	/* Free the enumerator and udev objects. */
 	udev_enumerate_unref(enumerate);
 	udev_unref(udev);
 
+	if (root == NULL) {
+		if (vendor_id == 0 && product_id == 0) {
+			register_global_error("No HID devices found in the system.");
+		} else {
+			register_global_error("No HID devices with requested VID/PID found in the system.");
+		}
+	}
+
 	return root;
 }
 
@@ -868,14 +922,17 @@ void  HID_API_EXPORT hid_free_enumeration(struct hid_device_info *devs)
 
 hid_device * hid_open(unsigned short vendor_id, unsigned short product_id, const wchar_t *serial_number)
 {
-	/* Set global error to none */
-	register_global_error(NULL);
-
 	struct hid_device_info *devs, *cur_dev;
 	const char *path_to_open = NULL;
 	hid_device *handle = NULL;
 
+	/* register_global_error: global error is reset by hid_enumerate/hid_init */
 	devs = hid_enumerate(vendor_id, product_id);
+	if (devs == NULL) {
+		/* register_global_error: global error is already set by hid_enumerate */
+		return NULL;
+	}
+
 	cur_dev = devs;
 	while (cur_dev) {
 		if (cur_dev->vendor_id == vendor_id &&
@@ -898,7 +955,7 @@ hid_device * hid_open(unsigned short vendor_id, unsigned short product_id, const
 		/* Open the device */
 		handle = hid_open_path(path_to_open);
 	} else {
-		register_global_error("No such device");
+		register_global_error("Device with requested VID/PID/(SerialNumber) not found");
 	}
 
 	hid_free_enumeration(devs);
@@ -908,52 +965,36 @@ hid_device * hid_open(unsigned short vendor_id, unsigned short product_id, const
 
 hid_device * HID_API_EXPORT hid_open_path(const char *path)
 {
-	/* Set global error to none */
-	register_global_error(NULL);
-
 	hid_device *dev = NULL;
 
 	hid_init();
+	/* register_global_error: global error is reset by hid_init */
 
 	dev = new_hid_device();
+	if (!dev) {
+		register_global_error("Couldn't allocate memory");
+		return NULL;
+	}
 
-	/* OPEN HERE */
-	dev->device_handle = open(path, O_RDWR);
+	dev->device_handle = open(path, O_RDWR | O_CLOEXEC);
 
-	/* If we have a good handle, return it. */
 	if (dev->device_handle >= 0) {
-		/* Set device error to none */
-		register_device_error(dev, NULL);
-
-		/* Get the report descriptor */
 		int res, desc_size = 0;
-		struct hidraw_report_descriptor rpt_desc;
 
-		memset(&rpt_desc, 0x0, sizeof(rpt_desc));
-
-		/* Get Report Descriptor Size */
+		/* Make sure this is a HIDRAW device - responds to HIDIOCGRDESCSIZE */
 		res = ioctl(dev->device_handle, HIDIOCGRDESCSIZE, &desc_size);
-		if (res < 0)
-			register_device_error_format(dev, "ioctl (GRDESCSIZE): %s", strerror(errno));
-
-		/* Get Report Descriptor */
-		rpt_desc.size = desc_size;
-		res = ioctl(dev->device_handle, HIDIOCGRDESC, &rpt_desc);
 		if (res < 0) {
-			register_device_error_format(dev, "ioctl (GRDESC): %s", strerror(errno));
-		} else {
-			/* Determine if this device uses numbered reports. */
-			dev->uses_numbered_reports =
-				uses_numbered_reports(rpt_desc.value,
-				                      rpt_desc.size);
+			hid_close(dev);
+			register_device_error_format(dev, "ioctl(GRDESCSIZE) error for '%s', not a HIDRAW device?: %s", path, strerror(errno));
+			return NULL;
 		}
 
 		return dev;
 	}
 	else {
-		/* Unable to open any devices. */
-		register_global_error(strerror(errno));
+		/* Unable to open a device. */
 		free(dev);
+		register_global_error_format("Failed to open a device with path '%s': %s", path, strerror(errno));
 		return NULL;
 	}
 }
@@ -1010,9 +1051,11 @@ int HID_API_EXPORT hid_read_timeout(hid_device *dev, unsigned char *data, size_t
 		else {
 			/* Check for errors on the file descriptor. This will
 			   indicate a device disconnection. */
-			if (fds.revents & (POLLERR | POLLHUP | POLLNVAL))
+			if (fds.revents & (POLLERR | POLLHUP | POLLNVAL)) {
 				// We cannot use strerror() here as no -1 was returned from poll().
+				register_device_error(dev, "hid_read_timeout: unexpected poll error (device disconnected)");
 				return -1;
+			}
 		}
 	}
 
@@ -1047,6 +1090,8 @@ int HID_API_EXPORT hid_send_feature_report(hid_device *dev, const unsigned char
 {
 	int res;
 
+	register_device_error(dev, NULL);
+
 	res = ioctl(dev->device_handle, HIDIOCSFEATURE(length), data);
 	if (res < 0)
 		register_device_error_format(dev, "ioctl (SFEATURE): %s", strerror(errno));
@@ -1058,6 +1103,8 @@ int HID_API_EXPORT hid_get_feature_report(hid_device *dev, unsigned char *data,
 {
 	int res;
 
+	register_device_error(dev, NULL);
+
 	res = ioctl(dev->device_handle, HIDIOCGFEATURE(length), data);
 	if (res < 0)
 		register_device_error_format(dev, "ioctl (GFEATURE): %s", strerror(errno));
@@ -1065,13 +1112,17 @@ int HID_API_EXPORT hid_get_feature_report(hid_device *dev, unsigned char *data,
 	return res;
 }
 
-// Not supported by Linux HidRaw yet
 int HID_API_EXPORT HID_API_CALL hid_get_input_report(hid_device *dev, unsigned char *data, size_t length)
 {
-	(void)dev;
-	(void)data;
-	(void)length;
-	return -1;
+	int res;
+
+	register_device_error(dev, NULL);
+
+	res = ioctl(dev->device_handle, HIDIOCGINPUT(length), data);
+	if (res < 0)
+		register_device_error_format(dev, "ioctl (GINPUT): %s", strerror(errno));
+
+	return res;
 }
 
 void HID_API_EXPORT hid_close(hid_device *dev)
@@ -1079,38 +1130,108 @@ void HID_API_EXPORT hid_close(hid_device *dev)
 	if (!dev)
 		return;
 
-	int ret = close(dev->device_handle);
-
-	register_global_error((ret == -1)? strerror(errno): NULL);
+	close(dev->device_handle);
 
 	/* Free the device error message */
 	register_device_error(dev, NULL);
 
+	hid_free_enumeration(dev->device_info);
+
 	free(dev);
 }
 
 
 int HID_API_EXPORT_CALL hid_get_manufacturer_string(hid_device *dev, wchar_t *string, size_t maxlen)
 {
-	return get_device_string(dev, DEVICE_STRING_MANUFACTURER, string, maxlen);
+	if (!string || !maxlen) {
+		register_device_error(dev, "Zero buffer/length");
+		return -1;
+	}
+
+	struct hid_device_info *info = hid_get_device_info(dev);
+	if (!info) {
+		// hid_get_device_info will have set an error already
+		return -1;
+	}
+
+	if (info->manufacturer_string) {
+		wcsncpy(string, info->manufacturer_string, maxlen);
+		string[maxlen - 1] = L'\0';
+	}
+	else {
+		string[0] = L'\0';
+	}
+
+	return 0;
 }
 
 int HID_API_EXPORT_CALL hid_get_product_string(hid_device *dev, wchar_t *string, size_t maxlen)
 {
-	return get_device_string(dev, DEVICE_STRING_PRODUCT, string, maxlen);
+	if (!string || !maxlen) {
+		register_device_error(dev, "Zero buffer/length");
+		return -1;
+	}
+
+	struct hid_device_info *info = hid_get_device_info(dev);
+	if (!info) {
+		// hid_get_device_info will have set an error already
+		return -1;
+	}
+
+	if (info->product_string) {
+		wcsncpy(string, info->product_string, maxlen);
+		string[maxlen - 1] = L'\0';
+	}
+	else {
+		string[0] = L'\0';
+	}
+
+	return 0;
 }
 
 int HID_API_EXPORT_CALL hid_get_serial_number_string(hid_device *dev, wchar_t *string, size_t maxlen)
 {
-	return get_device_string(dev, DEVICE_STRING_SERIAL, string, maxlen);
+	if (!string || !maxlen) {
+		register_device_error(dev, "Zero buffer/length");
+		return -1;
+	}
+
+	struct hid_device_info *info = hid_get_device_info(dev);
+	if (!info) {
+		// hid_get_device_info will have set an error already
+		return -1;
+	}
+
+	if (info->serial_number) {
+		wcsncpy(string, info->serial_number, maxlen);
+		string[maxlen - 1] = L'\0';
+	}
+	else {
+		string[0] = L'\0';
+	}
+
+	return 0;
+}
+
+
+HID_API_EXPORT struct hid_device_info *HID_API_CALL hid_get_device_info(hid_device *dev) {
+	if (!dev->device_info) {
+		// Lazy initialize device_info
+		dev->device_info = create_device_info_for_hid_device(dev);
+	}
+
+	// create_device_info_for_hid_device will set an error if needed
+	return dev->device_info;
 }
 
 int HID_API_EXPORT_CALL hid_get_indexed_string(hid_device *dev, int string_index, wchar_t *string, size_t maxlen)
 {
-	(void)dev;
 	(void)string_index;
 	(void)string;
 	(void)maxlen;
+
+	register_device_error(dev, "hid_get_indexed_string: not supported by hidraw");
+
 	return -1;
 }
 
